"""
This file implements policy-gradient based agents.
These agents range from vanila policy gradient (aka REINFORCE)[1] to
proximal policy optimization [2].

For specifications of individual loss functions, see graphity.nn.update_rules

[1] Policy Gradient Methods for Reinforcement Learning with Function Approximation. Sutton et al.
[2] Proximal Policy Optimization Algorithms, Schulman et al.
"""
import copy
import types

import torch
import torch.nn as nn
import torch.optim

import librl.agent
import librl.nn.pg_loss

# It caches the last generated policy in self.policy_latest, which can be sampled for additional actions.
@librl.agent.add_agent_attr(allow_update=True, policy_based=True)
class REINFORCEAgent(nn.Module):
    def __init__(self, actor_net, gamma=None, beta=None, alpha=None, l2=None, actor_loss_mult=None):
        super(REINFORCEAgent, self).__init__()
        self.alpha = alpha if alpha else self.get_default_hyperparameters().alpha
        self.l2 = l2 if l2 else self.get_default_hyperparameters().l2
        self.actor_loss_mult = actor_loss_mult if actor_loss_mult else self.get_default_hyperparameters().actor_loss_mult
        
        # Cache the last generated policy, so that we can sample for additional actions.
        self.policy_latest = None
        self.actor_net = actor_net
        # Don't overwrite defaults of VPG if we are going to be passing in a None
        self._actor_loss = librl.nn.pg_loss.VPG(**{k:v for (k,v) in {'gamma':gamma, 'beta':beta}.items() if v is not None})
        self.actor_optimizer = torch.optim.Adam(self.actor_net.parameters(), lr=self.alpha, weight_decay=self.l2)

    @staticmethod
    def get_default_hyperparameters():
        ret = {}
        ret['alpha'] = .001
        ret['l2'] = 0.1
        ret['actor_loss_mult'] = -1
        return types.SimpleNamespace(**ret)

    def act(self, state):
        return self(state)

    def forward(self, state):
        # Don't return policy information, so as to conform with stochastic agents API.
        actions, logprobs, self.policy_latest = self.actor_net(state)
        return actions, logprobs

    def actor_loss(self, task):
        return self._actor_loss(task)

    def steal(self):
        return copy.deepcopy(self.state_dict())

    def stuff(self, params):
        self.load_state_dict(params, strict=True)

# Implement a common framework for all synchronous actor-critic methods.
# It achieves this versatility by allowing you to specify the policy loss
# function, enabling policy gradient with baseline and PPO to use the same agent.
# You, the user, are responsible for supplying a policy network and value network
# that make sense for the problem.
# It caches the last generated policy in self.policy_latest, which can be sampled for additional actions.
@librl.agent.add_agent_attr(allow_update=True, policy_based=True)
class ActorCriticAgent(nn.Module):
    def __init__(self, critic_net, actor_net, actor_loss, actor_alpha=None, actor_l2=None,
                 critic_alpha=None, critic_l2=None, critic_steps=None, actor_loss_mult=None):
        super(ActorCriticAgent, self).__init__()
        self.actor_loss_mult = actor_loss_mult if actor_loss_mult else self.get_default_hyperparameters().actor_loss_mult
        self.actor_alpha = actor_alpha if actor_alpha else self.get_default_hyperparameters().actor_alpha
        self.actor_l2 = actor_l2 if actor_l2 else self.get_default_hyperparameters().actor_l2
        self.critic_alpha = critic_alpha if critic_alpha else self.get_default_hyperparameters().critic_alpha
        self.critic_l2 = critic_l2 if critic_l2 else self.get_default_hyperparameters().critic_l2
        self.critic_steps = critic_steps if critic_steps else self.get_default_hyperparameters().critic_steps


        self.policy_latest = None
        self.actor_net = actor_net
        self._actor_loss = actor_loss
        # TODO: Optimize with something other than ADAM.
        self.actor_optimizer = torch.optim.Adam(self.actor_net.parameters(), lr=self.actor_alpha, weight_decay=self.actor_l2)

        # Trust that the caller gave a reasonable value network.
        self.critic_net = critic_net
        # Goal of our value network (aka the critic) is to make our actual and expected values be equal.
        self._critic_loss = torch.nn.MSELoss(reduction="mean")
        # TODO: Optimize with something other than ADAM.
        self.critic_optimizer = torch.optim.Adam(self.critic_net.parameters(), lr=self.critic_alpha, weight_decay=self.critic_l2)

    @staticmethod
    def get_default_hyperparameters():
        ret = {}
        ret['actor_loss_mult'] = -1
        ret['actor_alpha'] = .001
        ret['actor_l2'] = 0.1
        ret['critic_alpha'] = .001
        ret['critic_l2'] = 0.1
        ret['critic_steps'] = 10
        return types.SimpleNamespace(**ret)

    def act(self, state):
        return self(state)

    def value(self, state):
        return self.critic_net(state)

    def forward(self, state):
        # Don't return policy information, so as to conform with stochastic agents API.
        actions, logprobs, self.policy_latest = self.actor_net(state)
        return actions, logprobs
        
    def actor_loss(self, task):
        return self._actor_loss(task)

    def critic_loss(self, task):
        # Must reshape states to be a batched 1d array.
        # TODO: Will need different reshaping for CNN's.
        losses = []
        for trajectory in task.trajectories:
            states = trajectory.state_buffer[:trajectory.done]
            states = states.view(-1, *(states.shape[1:]))
            losses.append(self._critic_loss(self.critic_net(states), trajectory.reward_buffer.view(-1, 1)[:trajectory.done]))
        return sum(losses)

    def steal(self):
        return copy.deepcopy(self.state_dict())
        
    def stuff(self, params):
        self.load_state_dict(params, strict=True)



import torch

# Get hyperparameters that are common between all agents/environments/tasks.
def get_default_hyperparams():
    ret = {}
    ret['allow_cuda'] = True
    ret['epochs'] = 10
    ret['episode_count'] = 1
    ret['episode_length'] = 10000
    # Neural net configurations
    ret['critic_steps'] = 10
    ret['dropout'] = .1
    ret['l2'] = .1
    ret['alpha'] = .0003
    # RL configurations
    ret['gamma'] = .95
    return ret


def to_cuda(tensor, allow_cuda):
    if allow_cuda and torch.cuda.is_available():
        return tensor.to(device="cuda")
    else:
        return tensor


import torch

# Given a buffer of actions and policies, compute the log probability of choosing action[t+1] under policy[t]
def old_log_probs(action_buffer, policy_buffer):
	logprob_old = torch.zeros([*policy_buffer.shape], dtype=torch.float32, device=action_buffer.device)
	max_t = policy_buffer.shape[-1] - 1
	# TODO: Figure out what the bootstrap value should be
	logprob_old[max_t] = policy_buffer[max_t].log_prob(action_buffer[max_t]).sum()
	for t in range(max_t):
		logprob_old[t+1] = policy_buffer[t].log_prob(action_buffer[t+1]).sum()
	return logprob_old

# Compute the generalized advantage estimation of our model.
# Caller must compute estimated values outside of this function.
def gae(reward_buffer, estimated_values, gamma, lambd=1):
	advantage = torch.zeros([*reward_buffer.shape], dtype=torch.float32, device=reward_buffer.device)
	max_t = reward_buffer.shape[-1] - 1
	# TODO: Determine how to predict "last" advantage value
	previous, advantage[max_t] = 0, estimated_values[max_t] -  reward_buffer[max_t]
	for t in range(max_t - 1, -1, -1):
		reward = (reward_buffer[t]
				+ gamma*estimated_values[t+1] - estimated_values[t])
		previous = advantage[t] = reward + lambd*gamma*previous
	return advantage

# Compute (in)finite horizion discounted rewards.
# Uses discount factor of gamma, and multiplies lambd(a) by the current element.
# May cause excessively deep gradients, but computes in linear time.	
def discounted_returns(reward_buffer, gamma):
	assert 0.0 < gamma and gamma <= 1.0
	discounted_rewards = torch.zeros([*reward_buffer.shape], dtype=torch.float32, device=reward_buffer.device)
	
	# By iterating backwards over the rewards, we are computing a linear filter on the output.
	# That is: O[t] = I[t] + gamma*O[t+1], where I is the input array and O is the output array.
	# This works by "splitting off" the first term of the summation at t'=t, and then reusing the already-computer sum
	# from t'=t+1..T-1.
	# This approach may accumulate larger roundoff errors when t ≪ stop_time.
	previous = 0
	for t in range(reward_buffer.shape[0]-1, -1, -1):
		previous =  discounted_rewards[t] = reward_buffer[t] + gamma*previous
	return discounted_rewards



import functools

import torch

import librl.replay
from .maml import maml_meta_step, compute_meta_loss, step_meta_optimizer
from .pg import policy_gradient_step, policy_gradient_update
def update_unshared_cc_agent(agent, tasks, it, loss_fn, loss_mul, optim):
    for i in range(it):
        # Iterate across all tasks, accumulating losses using the current task's agent's loss function.
        # Must init reduce with 0 or else first x is an task, not a scalar.
        # Provide lambda returning 0 in case the specified loss function DNE.
        losses = functools.reduce(lambda x,y: x + getattr(agent, loss_fn, lambda _: 0)(y), tasks, 0)
        # Compute mean and apply optional loss multiplier.
        losses *= (loss_mul / functools.reduce(lambda x,y: x + y.trajectory_count, tasks, 0))
        losses.backward()

        torch.nn.utils.clip_grad_norm_(agent.parameters(), 5)
        getattr(agent, optim).step(), agent.zero_grad()


import functools
import types

import more_itertools
import torch
import torch.nn as nn
import torch.optim

# Agent network based on a submission to my COSC689 class
# It is a stochastic policy network. It will return the policy from forward,
# and you can use this policy to generate further samples
# The current policy is sampling random values from a torch Categorical distrubtion
# conditioned on the output of a linear network.
# The Categorical distribution is non-differentiable, so this may cause
# problems for future programmers.
class MLPKernel(nn.Module):
    def __init__(self, input_dimensions, layer_list=None, dropout=None):
        
        super(MLPKernel, self).__init__()
        dropout = dropout if dropout else self.get_default_hyperparameters().dropout
        layer_list = layer_list if layer_list else self.get_default_hyperparameters().layer_list
        self.input_dimensions = list(more_itertools.always_iterable(input_dimensions))

        # Build linear layers from input defnition.
        linear_layers = []
        previous = functools.reduce(lambda x, y: x*y, self.input_dimensions, 1)
        for index,layer in enumerate(layer_list):
            linear_layers.append(nn.Linear(previous, layer))
            linear_layers.append(nn.LeakyReLU())
            # We have an extra component at the end, so we can dropout after every layer.
            linear_layers.append(nn.Dropout(dropout))
            previous = layer
        self.output_dimension = (previous, )
        self.linear_layers = nn.Sequential(*linear_layers)

        # Initialize NN
        for x in self.parameters():
            if x.dim() > 1:
                nn.init.kaiming_normal_(x)

    @staticmethod
    def get_default_hyperparameters():
        ret = {}
        ret['dropout'] = .1
        ret['layer_list'] = [200, 100]
        return types.SimpleNamespace(**ret)

    def forward(self, input):
        input = input.view(-1, functools.reduce(lambda x, y: x*y, self.input_dimensions, 1))
        # Push observations through feed forward layers.
        output = self.linear_layers(input.float())

        assert not torch.isnan(output).any()

        return output



import torch
import numpy as np

import librl.utils

# Enapsulate all replay memory of a single task
# TODO: Allow data to be moved to GPU, and make device required.
class Episode:
	def __init__(self, obs_space, act_space, episode_length=200, device='cpu'):
		self.state_buffer = torch.zeros([episode_length, *obs_space.shape], dtype=librl.utils.convert_np_torch(obs_space.dtype)).to(device)
		self.action_buffer = torch.zeros([episode_length, *act_space.shape], dtype=librl.utils.convert_np_torch(act_space.dtype)).to(device)
		self.logprob_buffer = torch.zeros([episode_length], dtype=torch.float32).to(device)
		self.reward_buffer = torch.zeros([episode_length], dtype=torch.float32).to(device)
		self.policy_buffer = np.full([episode_length], None, dtype=object)
		self.done =  None

	def log_state(self, t, state):
		self.state_buffer[t] = state
	def log_action(self, t, action, logprob):
		self.action_buffer[t] = action
		self.logprob_buffer[t] = logprob
	def log_rewards(self, t, reward):
		self.reward_buffer[t] = reward
	def log_policy(self, t, policy):
		self.policy_buffer[t]= policy
	def log_done(self, t):
		self.done = t

	def clear_replay(self):
		map(lambda x: x.fill_(0).detach_(), [self.state_buffer, self.action_buffer, self.logprob_buffer, self.reward_buffer])
		self.policy_buffer.fill(None)
		self.done = None


import math

import torch
import torch.distributions, torch.nn.init

#########################
#     Policy Objects    #
#########################
# Encapsulate the sampling / logprobs of a policy within a class.
# This helps out the VPG/PPO/PGB code because it doesn't need to know the form
# of the action distribution, it just has to call .log_prob()

class BiCategoricalPolicy:
    def __init__(self, first_seed, second_seed):
        self.first_seed = first_seed
        self.second_seed = second_seed
        # Create integer distributions from my network outputs
        self.first_dist = torch.distributions.Categorical(first_seed)
        self.second_dist = torch.distributions.Categorical(second_seed)

    def log_prob(self, actions):
        # Since I'm dealing log probs rather than probs, must add together.
        return self.first_dist.log_prob(actions[:,0]) + self.second_dist.log_prob(actions[:,1])

    def sample(self, size):
        # Sample edge pair to toggle
        first_sample = self.first_dist.sample(size)
        second_sample = self.first_dist.sample(size)
        return torch.stack([first_sample, second_sample], dim=-1)

# Action distribution based on a multivariate-normal distribution.
# Allows actions to be cross-correlated
class MultivariateNormal:
    def __init__(self, mu, lower_tril):
        self.mu = mu
        self.lower_tril = lower_tril
        # Create integer distributions from my network outputs
        self.dist = torch.distributions.MultivariateNormal(mu, scale_tril=self.lower_tril)

    def log_prob(self, actions):
        # Since I'm dealing log probs rather than probs, must add together.
        return self.dist.log_prob(actions)

    def sample(self, size):
        return self.dist.rsample(size)

# Action distribution based on repeated, independent normal distributions.
# Requires that all actions are independent.
class RepeatedNormal:
    def __init__(self, mu, var):
        self.mu = mu
        self.std = var.sqrt()
        # Create integer distributions from my network outputs
        self.dist = torch.distributions.Normal(mu, self.std)

    def log_prob(self, actions):
        # Since I'm dealing log probs rather than probs, must add together.
        return torch.sum(self.dist.log_prob(actions))

    def sample(self, size):
        return self.dist.rsample(size)

# Credit to exisiting "cheating" normal distribution that seems to perform well.
# However, in doesn't seem to do much better than an actual normal distribution in pratice.
# See:
#    https://github.com/andrewliao11/pytorch-a3c-mujoco/blob/master/train.py
class CheatingNormal:
    def __init__(self, mu, var):
        self.mu = mu
        self.var = var

    def log_prob(self, actions):
        pi = torch.FloatTensor([math.pi])
        a = (-1*(actions-self.mu).pow(2)/(2*self.var)).exp()
        b = 1/(2*self.var*pi.expand_as(self.var)).sqrt()
        return -a*b

    def sample(self, size):
        return self.mu + self.var.sqrt()*torch.randn((*size, *self.mu.shape))


import gym
import torch

import librl.replay
# The current module will be imported before librl.task is finished, so we
# will need to refer to our parent class by relative path.
from .base import Task as _Task

class ContinuousControlTask(_Task):
    # Aliase batch size as the episode count.
    episode_count=property(_Task.batch_size, _Task.batch_size)

    # Methods to fill a task's sequential replay buffer.
    @staticmethod
    def sample_trajectories(task):
        task.clear_trajectories()
        task.init_env()
        for i in range(task.trajectory_count):
            state = torch.tensor(task.env.reset()).to(task.device)
            episode = librl.replay.Episode(task.env.observation_space, task.env.action_space, task.episode_length)
            episode.log_done(task.episode_length + 1)
            for t in range(task.episode_length):
                
                episode.log_state(t, state)

                action, logprob_action = task.agent.act(state)
                episode.log_action(t, action, logprob_action)
                if task.agent.policy_based: episode.log_policy(t, task.agent.policy_latest)
                x = action.view(-1).detach().cpu().numpy()
                state, reward, done, _ = task.env.step(x)
                state, reward = torch.tensor(state).to(task.device), torch.tensor(reward).to(task.device)

                episode.log_rewards(t, reward)
                if done: 
                    episode.log_done(t+1)
                    break

            task.add_trajectory(episode)

    # Use sample_trajectories as the default sample, unless otherwise specified.
    def __init__(self, sample_fn = sample_trajectories.__func__, env=None, agent=None, trajectories=1, episode_length=100):

        super(ContinuousControlTask,self).__init__(librl.task.ProblemTypes.ContinuousControl)
        assert env is not None and agent is not None
        assert isinstance(env, gym.Env)

        self.env = env
        self.agent = agent
        self._episode_length = episode_length
        self.trajectory_count = trajectories

        self.sample = sample_fn
        self.trajectories = []

    # Override in subclass!!
    def init_env(self):
        raise NotImplemented("Please implement this method in your subclass.")

    def add_trajectory(self, trajectory):
        self.trajectories.append(trajectory)
    def clear_trajectories(self):
        self.trajectories = []

    # Let the task control the sampled batch size
    @property
    def episode_length(self):
        return self._episode_length
    @episode_length.setter
    def episode_length(self, value):
        self._episode_length = value

# Example class that allows us to run gym tasks with no extra effort.
class ContinuousGymTask(ContinuousControlTask):
	def __init__(self, **kwargs):
		super(ContinuousGymTask, self).__init__(**kwargs)
	def init_env(self):
        # Most gym environment need no extra init'ing
		pass


import functools

import torch

def cc_episodic_trainer(train_info, task_dist, train_fn):
    for epoch in range(train_info['epochs']):
        task_samples = task_dist.sample(train_info['episode_count'])
        train_fn(task_samples)

        rewards, mu_act = len(task_samples) * [None],  len(task_samples) * [None]
        for idx, task in enumerate(task_samples):
            mu_act[idx] = torch.mean(task.trajectories[0].action_buffer.type(torch.float32), (0))
            rewards[idx] = sum(task.trajectories[0].reward_buffer.view(-1))

        mean_reward = (sum(rewards)/len(rewards)).item()
        mean_action = functools.reduce(lambda x, y: x+y, mu_act, 0).mean()
        max_action = functools.reduce(lambda x, y: torch.max(x.abs(), y.abs()), mu_act)
        print(f"R^bar_({epoch}) = {mean_reward} with {mean_action:.4f} {max_action.data}.")



import functools

import librl.agent.pg, librl.agent.mdp
import librl.task
import librl.train.cc

@functools.singledispatch
def compute_meta_loss(agent, task):
    # If you introduce a new agent type, and you wish to allow this agent to be updated via
    # MAML, you must register an override for this function.
    # Must return an iterable of losses.
    raise NotImplemented(f"Meta loss computation not implemented for this agent type.")

@compute_meta_loss.register(librl.agent.mdp.RandomAgent)
def _(agent, task):
    return []

@compute_meta_loss.register(librl.agent.pg.REINFORCEAgent)
def _(agent, task):
    return (getattr(agent, 'actor_loss_mult', 1) * agent.actor_loss(task),)

@compute_meta_loss.register(librl.agent.pg.ActorCriticAgent)
def _(agent, task):
    return (agent.critic_loss(task), getattr(agent,'actor_loss_mult', 1) * agent.actor_loss(task))



@functools.singledispatch
def step_meta_optimizer(agent):
        # If you introduce a new agent type, and you wish to allow this agent to be updated via
    # MAML, you must register an override for this function.
    raise NotImplemented(f"Meta optimizer step not implemented for this agent type.")

@step_meta_optimizer.register(librl.agent.mdp.RandomAgent)
def _(agent):
    pass
@step_meta_optimizer.register(librl.agent.pg.REINFORCEAgent)
def _(agent):
    # TODO: Grab grad clip from agent's hypers.
    #torch.nn.utils.clip_grad_norm(agent.parameters(), 40)
    agent.actor_optimizer.step()
@step_meta_optimizer.register(librl.agent.pg.ActorCriticAgent)
def _(agent):
    # TODO: Grab grad clip from agent's hypers.
    #torch.nn.utils.clip_grad_norm(agent.parameters(), 40)
    agent.actor_optimizer.step()
    agent.critic_optimizer.step()


# Implement MAML algorithm for meta-RL tasks
# Assumes there is no interaction between different agents.
# If components of an agent are shared between tasks, gradient updates are no longer independent.
def maml_meta_step(task_samples, adapt_steps = 1):
    for task in task_samples: assert task.problem_type == librl.task.ProblemTypes.ContinuousControl
    # Collect all unique agents in our task list.
    agents = { (id(task.agent),task.agent) for task in task_samples}
    # Make sure we do not persist grads between training epochs.
    for _, agent in agents: agent.zero_grad()
    # Keep track of our starting parmeters.
    # Since we have potential redundant agents, only keep one copy of our params per agent.
    slow_parameters = {id_agent : agent.steal() for id_agent, agent in agents}
    # However, each task will have unique parameters after being adapted.
    size = len(task_samples)
    adapted_parameters, adapted_grads, adapted_rewards = size*[None], size*[None], size*[None]
    
    ####################
    # Task Adaptation  #
    ####################
    for idx, task in enumerate(task_samples):
        agent = task.agent
        agent.stuff(slow_parameters[id(agent)])
        # Perform an arbitrary number of adaptation steps
        for _ in range(adapt_steps): librl.train.cc.policy_gradient_step([task])
        
        # Sample a task for the meta-adaptation step and compute the loss based on agent type.
        task.sample(task)
        for loss in compute_meta_loss(agent, task): loss.backward()

        # All pytorch optimizers perform updates using grads.
        # If we copy out the grads and apply them later, pytorch won't be any wiser
        # as to how those grads got there. Use "clone" to prevent grads from being deleted.
        adapted_grads[idx] = [(p.grad.clone().detach() if p.grad != None else None)for p in agent.parameters()]
        adapted_rewards[idx] = functools.reduce(lambda x, y: x+sum(y.reward_buffer.view(-1)), task.trajectories, 0)

        # Prevent meta-grads from polluting the next task's update.
        agent.zero_grad()
        adapted_parameters[idx] = agent.steal()

    ####################
    # Meta-update step #
    ####################
    # We want to apply gradient updates to our starting weights,
    # and we want to discard any existing gradient updates
    for id_agent, agent in agents: agent.stuff(slow_parameters[id_agent]), agent.zero_grad()

    # Apply the i'th task's gradient update to the i'th task's agent.
    for idx, task in enumerate(task_samples):
        agent = task.agent
        for param, grad in zip(agent.parameters(), adapted_grads[idx]):
            # Some parts of a model may be frozen, so we can skip those params.
            if not param.requires_grad or grad is None: continue
            # If our grad object doesn't exist, we must create one in place.
            if param.grad is None: param.grad = grad
            # Otherwise accumulate grads in place.
            else: param.grad.add_(grad)
    
    # And apply those using our existing optimizer.
    for _, agent in agents: step_meta_optimizer(agent)




import numpy as np
from numpy.random import Generator, PCG64

# Class representing a weighted task distribution.
# You can sample a number of tasks from this with replacements
class TaskDistribution:
	def __init__(self, seed=None):
		self._tasks = []
		self._task_weights = []
		self._rng = Generator(PCG64(seed))

	def add_task(self, task, weight=1):
		self._tasks.append(task)
		self._task_weights.append(weight)

	def sample(self, count):
		# Must convert weights to proabbilities.
		mean = sum(self._task_weights)
		task_probabilities = list(map(lambda x: x/mean, self._task_weights))
		# Task definitions are different than task objects.
		# If they were the same, and we sampled on tasks directly,
		# it was possible to return the same task object twice.
		# This lead to accidentally overwriting replay memory.
		definitions = self._rng.choice(self._tasks, size=(count,), replace=True, p=task_probabilities)
		return [definition.instance() for definition in definitions]





import numpy as np
import torch

def convert_np_torch(item):
	if item in convert_np_torch._info: return convert_np_torch._info[item]
	elif item == np.float32: return torch.float32
	elif item == np.int8: return torch.int8
	print(f"Dtype ({item}) not implemented.")
	assert 0 and "Dtype not implemented."

convert_np_torch._info = {
		np.bool       : torch.bool,
		np.uint8      : torch.uint8,
		np.int8       : torch.int8,
		np.int16      : torch.int16,
		np.int32      : torch.int32,
		np.int64      : torch.int64,
		np.float16    : torch.float16,
		np.float32    : torch.float32,
		np.float64    : torch.float64,
		np.complex64  : torch.complex64,
		np.complex128 : torch.complex128}




from distutils.core import setup

setup(name='librl',
      version='0.1.2',
      description='Library for various DRL and DL projects.',
      author='Matthew McRaven',
      author_email='mkm302@georgetown.edu',
      url='https://github.com/Matthew-McRaven/librl',
      packages=['librl',
      'librl.agent',
      'librl.nn',
      'librl.replay',  
      'librl.task',
      'librl.train', 'librl.train.cc']
     )


from os import stat
import types

import torch
import torch.distributions, torch.nn.init
import torch.optim
from overrides import overrides, EnforceOverrides

import librl.calc
import librl.task
##########################
# Policy Gradient Losses #
##########################

# Implements a framework for policy gradient losses that use  entropy bonuses when computing discounted rewards.
class PolicyLossWithEntropyBonus(EnforceOverrides):

    def __init__(self, gamma, beta):
        self.gamma = gamma
        self.beta = beta

    # Compute discounted rewards with entropy bonus
    def get_discounted_rewards(self, trajectory):
        entropy = -self.beta * (trajectory.logprob_buffer[:trajectory.done]) * (trajectory.logprob_buffer[:trajectory.done].exp())
        bonus_rewards = trajectory.reward_buffer[:trajectory.done] + entropy
        return librl.calc.discounted_returns(bonus_rewards, gamma=self.gamma)

    # Override this method to compute the loss of a single trajectory.
    # Should return a single value.
    def compute_trajectory_loss(self, trajectory):
        raise NotImplemented("Must implement this in subclass")

    # When called with a CC task, run the loss algorithm over each trajectory in the task.
    # This will implicitly perform the outer mean over the the number of trajectories,
    # and defer the computation of per-trajectory loss to a subclass via compute_trajectory_loss.
    def __call__(self, task):
        assert isinstance(task, librl.task.ContinuousControlTask)
        losses = []
        for trajectory in task.trajectories: losses.append(self.compute_trajectory_loss(trajectory))
        return sum(losses) / len(losses)

# Vanilla policy gradient update / loss function.
class VPG(PolicyLossWithEntropyBonus):
    def __init__(self, gamma=.95, beta=0.01):
        super(VPG, self).__init__(gamma, beta)
        
    @overrides
    def compute_trajectory_loss(self, trajectory):
        return sum(trajectory.logprob_buffer[:trajectory.done] * self.get_discounted_rewards(trajectory))
       
# Policy gradient with baseline update / loss function.
class PGB(PolicyLossWithEntropyBonus):
    def __init__(self, critic_net,  gamma=.95, beta=.01):
        super(PGB, self).__init__(gamma, beta)
        self.critic_net = critic_net

    @overrides
    def compute_trajectory_loss(self, trajectory):
        # Don't propogate gradients into critic when updating actor.
        with torch.no_grad(): estimated_values = self.critic_net(trajectory.state_buffer).view(-1)[:trajectory.done]
        return sum(trajectory.logprob_buffer[:trajectory.done] * (self.get_discounted_rewards(trajectory)-estimated_values))

# Proximal policy optimization update / loss function.
class PPO(PolicyLossWithEntropyBonus):

    def __init__(self, critic_net, gamma=0.975, beta=0.01, lambd=.99, epsilon=.5, c_1=1):
        super(PPO, self).__init__(gamma, beta)
        self.critic_net = critic_net
        self.lambd = lambd
        self.epsilon = epsilon
        self.c_1 = c_1

    @overrides
    def compute_trajectory_loss(self, trajectory):
        # Don't propogate gradients into critic when updating actor.
        with torch.no_grad(): estimated_values = self.critic_net(trajectory.state_buffer).view(-1)[:trajectory.done]
        discounted = self.get_discounted_rewards(trajectory)
        A =  librl.calc.gae(trajectory.reward_buffer[:trajectory.done], estimated_values, self.gamma)
        log_prob_old = librl.calc.old_log_probs(trajectory.action_buffer[:trajectory.done], trajectory.policy_buffer[:trajectory.done])
        # Compute indiviudal terms of the PPO algorithm.
        ratio = trajectory.logprob_buffer[:trajectory.done].exp() / (log_prob_old.exp() + 1e-6)
        lhs, rhs = ratio * A, torch.clamp(ratio, 1-self.epsilon, 1+self.epsilon) * A
        minterm = torch.min(lhs, rhs)
        subterm = self.c_1 * (discounted-estimated_values.view(-1)).pow(2)
        return torch.sum(minterm - subterm)


import librl.task

# Describes a family of related tasks
# Single task object is not shared between episodes within an epoch.
# Otherwise, replay buffers will overwrite each other.
class Task():
    class Definition:
        def __init__(self, ctor, **kwargs):
            self.task_ctor = ctor
            self.task_kwargs = kwargs
        def instance(self):
            return self.task_ctor(**self.task_kwargs)
    def __init__(self, problem_type, device="cpu", batch_size=10):
        self._problem_type=problem_type
        self._device = device
        self._batch_size = batch_size
        
    # Make the problem type effectively constant
    @property
    def problem_type(self):
        return self._problem_type

    # Let the task control which device it wants to be scheduled on.
    @property
    def device(self):
        return self._device
    @device.setter
    def device(self, value):
        self._device = value

    # Let the task control the sampled batch size
    @property
    def batch_size(self):
        return self._batch_size
    @batch_size.setter
    def batch_size(self, value):
        self._batch_size = value